"""pfSense integration."""
import logging
from pydoc import cli
import time
from typing import Any

from homeassistant.components.update import (
    UpdateDeviceClass,
    UpdateEntity,
    UpdateEntityDescription,
)
from homeassistant.components.update.const import UpdateEntityFeature
from homeassistant.config_entries import ConfigEntry
from homeassistant.core import HomeAssistant, callback
from homeassistant.helpers import entity_platform
from homeassistant.helpers.entity import EntityCategory
from homeassistant.helpers.update_coordinator import DataUpdateCoordinator
from homeassistant.util import slugify

from . import CoordinatorEntityManager, OPNSenseEntity, dict_get
from .const import COORDINATOR, DOMAIN

_LOGGER = logging.getLogger(__name__)


async def async_setup_entry(
    hass: HomeAssistant,
    config_entry: ConfigEntry,
    async_add_entities: entity_platform.AddEntitiesCallback,
):
    """Set up the OPNSense update entities."""

    @callback
    def process_entities_callback(hass, config_entry):
        data = hass.data[DOMAIN][config_entry.entry_id]
        coordinator = data[COORDINATOR]
        entities = []
        entity = OPNSenseFirmwareUpdatesAvailableUpdate(
            config_entry,
            coordinator,
            UpdateEntityDescription(
                key=f"firmware.update_available",
                name="Firmware Updates Available",
                entity_category=EntityCategory.DIAGNOSTIC,
            ),
            True,
        )
        entities.append(entity)

        return entities

    cem = CoordinatorEntityManager(
        hass,
        hass.data[DOMAIN][config_entry.entry_id][COORDINATOR],
        config_entry,
        process_entities_callback,
        async_add_entities,
    )
    cem.process_entities()


class OPNSenseUpdate(OPNSenseEntity, UpdateEntity):
    def __init__(
        self,
        config_entry,
        coordinator: DataUpdateCoordinator,
        entity_description: UpdateEntityDescription,
        enabled_default: bool,
    ) -> None:
        """Initialize the sensor."""
        self.config_entry = config_entry
        self.entity_description = entity_description
        self.coordinator = coordinator
        self._attr_entity_registry_enabled_default = enabled_default
        self._attr_name = f"{self.opnsense_device_name} {entity_description.name}"
        self._attr_unique_id = slugify(
            f"{self.opnsense_device_unique_id}_{entity_description.key}"
        )

        self._attr_supported_features |= (
            UpdateEntityFeature.INSTALL
            # | UpdateEntityFeature.BACKUP
            # | UpdateEntityFeature.PROGRESS
            # | UpdateEntityFeature.RELEASE_NOTES
            # | UpdateEntityFeature.SPECIFIC_VERSION
        )

    @property
    def device_class(self):
        return UpdateDeviceClass.FIRMWARE


class OPNSenseFirmwareUpdatesAvailableUpdate(OPNSenseUpdate):
    @property
    def available(self):
        state = self.coordinator.data
        if state["firmware_update_info"] is None:
            return False

        try:
            status = state["firmware_update_info"]["status"]
            if status == "error":
                return False
        except:
            return False

        return super().available

    @property
    def title(self):
        return "OPNSense"

    @property
    def installed_version(self):
        """Version installed and in use."""
        state = self.coordinator.data

        try:
            return dict_get(state, "firmware_update_info.product.product_version")
        except KeyError:
            return None

    @property
    def latest_version(self):
        """Latest version available for install."""
        state = self.coordinator.data

        try:
            # fake a new update
            # return "foobar"
            product_version = dict_get(
                state, "firmware_update_info.product.product_version"
            )
            product_latest = dict_get(
                state, "firmware_update_info.product.product_latest"
            )
            if product_version is None or product_latest is None:
                return None

            if (
                dict_get(state, "firmware_update_info.status") == "update"
                and product_version == product_latest
            ):
                product_latest = product_latest + "+"

            return product_latest
        except KeyError:
            return None

    @property
    def in_progress(self):
        """Update installation in progress."""
        return False

    @property
    def extra_state_attributes(self):
        state = self.coordinator.data
        attrs = {}

        for key in [
            "status",
            "status_msg",
            "last_check",
            "os_version",
            "product_id",
            "product_target",
            "product_version",
            "upgrade_needs_reboot",
            "download_size",
        ]:
            slug_key = slugify(key)
            attrs[f"opnsense_{slug_key}"] = dict_get(
                state, f"firmware_update_info.{key}"
            )

        return attrs

    @property
    def release_url(self):
        return self.config_entry.data.get("url", None) + "/ui/core/firmware#changelog"

    @property
    def release_summary(self):
        try:
            state = self.coordinator.data

            if dict_get(state, "firmware_update_info.status") != "update":
                return None

            product_name = dict_get(state, "firmware_update_info.product.product_name")
            product_nickname = dict_get(
                state, "firmware_update_info.product.product_nickname"
            )
            product_version = dict_get(
                state, "firmware_update_info.product.product_version"
            )
            product_latest = dict_get(
                state, "firmware_update_info.product.product_latest"
            )
            status_msg = dict_get(state, "firmware_update_info.status_msg")
            upgrade_needs_reboot = dict_get(
                state, "firmware_update_info.upgrade_needs_reboot"
            )

            if upgrade_needs_reboot is None or upgrade_needs_reboot == "0":
                upgrade_needs_reboot = False

            if upgrade_needs_reboot == "1":
                upgrade_needs_reboot = True

            total_package_count = len(
                dict_get(state, "firmware_update_info.all_packages", {}).keys()
            )
            new_package_count = len(
                dict_get(state, "firmware_update_info.new_packages", [])
            )
            reinstall_package_count = len(
                dict_get(state, "firmware_update_info.reinstall_packages", [])
            )
            remove_package_count = len(
                dict_get(state, "firmware_update_info.remove_packages", [])
            )
            upgrade_package_count = len(
                dict_get(state, "firmware_update_info.upgrade_packages", [])
            )

            summary = """
## {} version {} ({})

{}

- reboot needed: {}
- total affected packages: {}
- new packages: {}
- reinstalled packages: {}
- removed packages: {}
- upgraded packages: {}
""".format(
                product_name,
                product_latest,
                product_nickname,
                status_msg,
                upgrade_needs_reboot,
                total_package_count,
                new_package_count,
                reinstall_package_count,
                remove_package_count,
                upgrade_package_count,
            )
        except:
            return None
        return summary

    def install(self, version=None, backup=False):
        """Install an update."""
        client = self._get_opnsense_client()
        task_details = client.upgrade_firmware()
        sleep_time = 10
        running = True
        while running:
            time.sleep(sleep_time)
            try:
                response = client.upgrade_status()
                # after finished status is "done"
                running = response["status"] == "running"
            except:
                pass

        # check needs_reboot, if yes trigger reboot
        response = client.get_firmware_update_info()
        # upgrade_needs_reboot = dict_get(response, "upgrade_needs_reboot")
        upgrade_needs_reboot = dict_get(response, "needs_reboot")

        if upgrade_needs_reboot is None or upgrade_needs_reboot == "0":
            upgrade_needs_reboot = False

        if upgrade_needs_reboot == "1":
            upgrade_needs_reboot = True

        if upgrade_needs_reboot:
            client.system_reboot()
